#include <assert.h>
#include <stdio.h>
#include <cuda.h>

#define ARRAY_SIZE  100
const int ARRAY_BYTES = ARRAY_SIZE * sizeof(int);

//kernel
__global__ void addKernel(  int  *device_a, int *device_b, int *device_result){
   int idx = threadIdx.x;
   device_result[idx] = device_a[idx] + device_b[idx];
}


void onDevice( int  *host_a, int *host_b, int *host_result ){

	int *device_a, *device_b, *device_result; 

   //Asigna la memoria en el dispositivo
	cudaMalloc( (void**)&device_a, ARRAY_BYTES );
	cudaMalloc( (void**)&device_b, ARRAY_BYTES );
	cudaMalloc( (void**)&device_result, ARRAY_BYTES );

    //Copia los arregloa 'a' y 'b' al dispositivo
	cudaMemcpy( device_a, host_a, ARRAY_BYTES, cudaMemcpyHostToDevice );
	cudaMemcpy( device_b, host_b, ARRAY_BYTES, cudaMemcpyHostToDevice );

    //Ejecuta el kernel
    addKernel<<<1,ARRAY_SIZE>>>( device_a, device_b, device_result);

    //Copia el resultado del dispositivo al host
	cudaMemcpy( host_result, device_result, ARRAY_BYTES, cudaMemcpyDeviceToHost );

    //Libera la memoria
	cudaFree( device_a );
	cudaFree( device_b );
	cudaFree( device_result );

}

void onHost(){

	int *host_a, *host_b, *host_result;
	
	//Asigna memoria en el host
	host_a = (int*)malloc(ARRAY_BYTES);
	host_b = (int*)malloc(ARRAY_BYTES);
	host_result = (int*)malloc(ARRAY_BYTES);

    //Llena los arreglos 'a' y 'b' en el CPU
    for (int i=0; i<ARRAY_SIZE; i++) {
        host_a[i] = i;
        host_b[i] = i*i;
		host_result[i]=0;
    }

    onDevice(host_a, host_b, host_result);

    //Comprobamos resultados
    for (int i=0; i<ARRAY_SIZE; i++) {
		assert( host_a[i] + host_b[i] == host_result[i] );
    }

	printf("\n (: Ejecucion correcta :) \n");
    
    // free host memory 
	free(host_a);
	free(host_b);
	free(host_result);
}

int main(){   
	onHost();
    return 0;
}

