#include <stdio.h>
#include "Error.h"

#define BLOCK_SIZE 3
#define STRIDE 6

__global__ void matrixKernel( float *d_in, float *d_out ){
    // Índice del bloque
    int bx = blockIdx.x;
    int by = blockIdx.y;
       
    // Índice del hilo (coeficiente actual)
    int tx = threadIdx.x;
    int ty = threadIdx.y;
                                              //i * STRIDE + j
    float sumando =  d_in[ (by * BLOCK_SIZE + 0) * STRIDE + (bx * BLOCK_SIZE + 0) ];
    float elemento =  d_in[ (by * BLOCK_SIZE + ty) * STRIDE + (bx * BLOCK_SIZE + tx) ];

    d_out[ (by * BLOCK_SIZE + ty) * STRIDE + (bx * BLOCK_SIZE + tx) ] = sumando + elemento;    
}

void onDevice( float *h_in, float *h_out, int ARRAY_BYTES){
    // Declaramos los apuntadores del GPU
    float *d_in;
    float *d_out;

    // Asignamos memoria en el GPU
    cudaMalloc( (void**)&d_in, ARRAY_BYTES) ;
    cudaMalloc( (void**)&d_out, ARRAY_BYTES) ;

    // Copiamos memoria del CPU al GPU
    cudaMemcpy( d_in, h_in, ARRAY_BYTES, cudaMemcpyHostToDevice );
    HANDLER_ERROR_MSG("memory H2D d_in");

    // Comfiguración de la ejecución
    dim3 ThreadsBlocks( 3, 3 );
    dim3 GridBlocks( 2 , 2);  

    matrixKernel<<< GridBlocks, ThreadsBlocks >>>( d_in, d_out );
    HANDLER_ERROR_MSG("kernel panic!!!");
    cudaThreadSynchronize();

    // Copiar los resultados del GPU al CPU
    cudaMemcpy( h_out, d_out, ARRAY_BYTES, cudaMemcpyDeviceToHost );
    HANDLER_ERROR_MSG("memory H2D d_out");

    // Liberamos memoria del GPU
    cudaFree( d_in );
    cudaFree( d_out );
}

void test(float *h_out, int ARRAY_SIZE){
    // Verificamos que el trabajo fue correcto
    for (int i=0, x=0, y=0; i<ARRAY_SIZE; i++, x++) {
            printf( " %.2f ",h_out[i] );
            if(x==(STRIDE-1)/2)
                printf("\t");
            if(x==STRIDE-1){
                printf("\n");
                x=-1;
                y++;
            }
            if(y==STRIDE/2){
                printf("\n");
                y++;
            }
    }
    printf("\n (: Ejecucion correcta :) \n");
}

void onHost(){
    const int ARRAY_SIZE = 36;
    const int ARRAY_BYTES = ARRAY_SIZE * sizeof(float);

    float h_in[] ={ 1.0, 2.0, 3.0,    10.0, 11.0, 12.0,
                    4.0, 5.0, 6.0,    13.0, 14.0, 15.0,
                    7.0, 8.0, 9.0,    16.0, 17.0, 18.0,
                    19.0, 20.0, 21.0, 28.0, 29.0, 30.0,
                    22.0, 23.0, 24.0, 31.0, 32.0, 33.0,
                    25.0, 26.0, 27.0, 34.0, 35.0, 36.0 };

    // Apuntador en el CPU
    float *h_out;

    // Asignamos memoria en el host
    h_out = (float*)malloc( ARRAY_BYTES );

    for(int i = 0; i < ARRAY_SIZE; i++){
        h_out[i] = 0;
    }

    // Llamada a la configuración del dispositivo
    onDevice( h_in, h_out, ARRAY_BYTES );
    test(h_out, ARRAY_SIZE);

    // Liberamos memoria del CPU
    free( h_out );
}


int main( void ) {  
    onHost();
    return 0;
}
