#include <stdio.h>
#include <assert.h>
#include "Error.h"

#define N 250

__global__ void transposedMatrixKernel(int *d_a, int *d_b){
	int i = threadIdx.x+blockIdx.x*blockDim.x;
    int j = threadIdx.y+blockIdx.y*blockDim.y;

	while( i < N ){
		j = threadIdx.y + blockIdx.y * blockDim.y;
		while( j < N ){
			d_b[ i*N+j ] = d_a[j*N+i];
			j += blockDim.y * gridDim.y;
		}
		i += blockDim.x * gridDim.x;
	}
}


void onDevice(int h_a[][N], int h_b[][N]  ){
	int *d_a, *d_b;

	const int ARRAY_BYTES = N * N * sizeof(int);

    // Asigna memoria en el GPU
	HANDLER_ERROR_ERR(cudaMalloc((void**)&d_a,ARRAY_BYTES));
	HANDLER_ERROR_ERR(cudaMalloc((void**)&d_b,ARRAY_BYTES));

    // Copiar memoria del CPU al GPU
	HANDLER_ERROR_ERR(cudaMemcpy(d_a, h_a, ARRAY_BYTES, cudaMemcpyHostToDevice));
	HANDLER_ERROR_ERR(cudaMemcpy(d_b, h_b, ARRAY_BYTES, cudaMemcpyHostToDevice));

    // Configuración de ejecución
	dim3 GridBlocks( 4,4 );
	dim3 ThreadsBlocks( 16,16 );
	
	transposedMatrixKernel<<<GridBlocks,ThreadsBlocks>>>( d_a, d_b );
    HANDLER_ERROR_MSG("kernel panic!!!");

    // Copiar de vuelta memoria del GPU al CPU
	HANDLER_ERROR_ERR(cudaMemcpy(h_b, d_b, ARRAY_BYTES, cudaMemcpyDeviceToHost));	

    // Libera memoria
	HANDLER_ERROR_ERR(cudaFree(d_a));
	HANDLER_ERROR_ERR(cudaFree(d_b));
}

void test(int h_a[][N], int h_b[][N] ){
	int i;
	int j;

	for(i=0; i < N; i++){
		for(j = 0; j < N; j++)
			assert(h_a[j][i] == h_b[i][j]);
	}
    printf("\n (: Ejecucion correcta :) \n");	
}

void onHost(){
	int i,j,k=0;
	int h_a[N][N], h_b[N][N];

	for(i = 0; i < N; i++){
		for(j = 0; j < N; j++){
			h_a[i][j] = k;
			h_b[i][j] = 0;
			k++;
		}
	}

	onDevice(h_a,h_b);
	test(h_a,h_b);
}

int main(){
	onHost();
}

